package pt.uc.dei.crf_ma;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.TreeMap;

import pt.uc.dei.crf_ma.types.TermExtractorInstance;
import cc.mallet.types.InstanceList;

public abstract class MTTrainer {
	protected int nthreads;
	protected String inputFolder; // folder holding the input data
	protected SequenceLabeller sequenceLabeller; // the CRF model
	protected int totalDataInstances = 0; // holds the total number of data instances for training in the input training data
	protected TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances; // holds the instances of the files of each tagger
	protected TreeMap<String, InstanceList> distinctInputInstances; // holds all the distinct instances found in the input folder, to determine ground truth, measure accuracy, sensitivity, etc.
	protected TreeMap<String, InstanceList> estimatedGroundTruth; // holds the current estimate of the model for the ground truth
	protected TreeMap<String, InstanceList> actualGroundTruth; // holds the actual ground truth of the data (when available)
	protected TreeMap<String, InstanceList> evaluationData; // holds the evaluation dataset (when available)
	protected PrintWriter pwStats = null; // file to store the stats of each iteration 
	
	public MTTrainer(String inputFolder, SequenceLabeller sequenceLabeller, int nthreads, TreeMap<String, TreeMap<String, InstanceList>> taggersDataInstances, int totalDataInstances, TreeMap<String, InstanceList> distinctInputInstances, TreeMap<String, InstanceList> actualGroundTruth, TreeMap<String, InstanceList> evaluationData) {
		this.nthreads = nthreads;
		this.inputFolder = inputFolder;
		this.sequenceLabeller = sequenceLabeller;
		this.totalDataInstances = totalDataInstances;
		this.taggersDataInstances = taggersDataInstances;
		this.distinctInputInstances = distinctInputInstances;
		this.actualGroundTruth = actualGroundTruth;
		this.evaluationData = evaluationData;
		try {
			this.pwStats = new PrintWriter("iteration_stats.tsv");
			this.pwStats.println("iteration	num_training_instances	num_iterations	trainingData_likelihood	sum_ll	avg_ll	" +
					"estimated_gt_tokenAcc	estimated_gt_precision	estimated_gt_recall	estimated_gt_f1	" +
					"avg_tagger_tokenAcc	avg_tagger_precision	avg_tagger_recall	avg_tagger_f1	" +
					"model_likelihood	avg_tagger_update");
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		}
	}

	public TreeMap<String, InstanceList> train(int maxIterations) {
		throw new UnsupportedOperationException();
	}
	
	public TreeMap<String, InstanceList> estimateGroundTruth(TreeMap<String, InstanceList> distinctInputInstances) {
		throw new UnsupportedOperationException();
	}
	
	// Tagger evaluation for keyphrase extraction task
//	public EvaluationResults evaluateTagger(String taggerName, TreeMap<String, InstanceList> taggerInstancesMap, TreeMap<String, InstanceList> estimatedGroundTruth) {
//		System.out.print("[MultipleTaggers] 	Evaluating tagger: " + taggerName);
//		int correctTokens = 0;
//		int numTokens = 0;
//		TreeMap<String, HashSet<String>> taggerKeywordsAllDocuments = new TreeMap<String, HashSet<String>>();
//		TreeMap<String, HashSet<String>> gtKeywordsAllDocuments = new TreeMap<String, HashSet<String>>();
//		for (String filename : taggerInstancesMap.keySet()) {
//			InstanceList taggerInstances = taggerInstancesMap.get(filename);
//			InstanceList groundTruthInstances = estimatedGroundTruth.get(filename);
//			HashSet<String> taggerKeywords = new HashSet<String>();
//			HashSet<String> gtKeywords = new HashSet<String>();
//			if (taggerInstances.size() != groundTruthInstances.size()) {
//				continue;
//			}
//			for (int i=0; i<groundTruthInstances.size(); i++) {
//				TermExtractorInstance taggerInstance = (TermExtractorInstance) taggerInstances.get(i);
//				TermExtractorInstance groundTruthInstance = (TermExtractorInstance) groundTruthInstances.get(i);
//				ArrayList<String> tokens = groundTruthInstance.getTokens();
//				ArrayList<String> taggerLabels = taggerInstance.getLabels();
//				ArrayList<String> groundTruthLabels = groundTruthInstance.getLabels();
//				if (taggerLabels.size() != groundTruthLabels.size()) {
//					continue;
//				}
//				for (int j=0; j<tokens.size(); j++) {
//					numTokens++;
//					String taggerLabel = taggerLabels.get(j);
//					String groundTruthLabel = groundTruthLabels.get(j);
//					if (taggerLabel.equals(groundTruthLabel)) {
//						correctTokens++;
//					} 
//				}
//				taggerKeywords = TermExtractionWithMultipleTaggers.getKeywords(taggerKeywords, tokens, taggerLabels);
//				gtKeywords = TermExtractionWithMultipleTaggers.getKeywords(gtKeywords, tokens, groundTruthLabels);
//			}
//			taggerKeywordsAllDocuments.put(filename, taggerKeywords);
//			gtKeywordsAllDocuments.put(filename, gtKeywords);
//		}
//		double tokenAccuracy = ((double) correctTokens) / ((double) numTokens);
//		EvaluationResults evaluationResults = TermExtractionWithMultipleTaggers.evaluateKeywords(taggerKeywordsAllDocuments, gtKeywordsAllDocuments);
//		evaluationResults.setNumTokens(numTokens);
//		evaluationResults.setTokenAccuracy(tokenAccuracy);
//		DecimalFormat f = new DecimalFormat("0.####");
//		System.out.println(" -> tokenAcc="+f.format(tokenAccuracy)+" precision="+f.format(evaluationResults.getPrecision())+
//				" recall="+f.format(evaluationResults.getRecall())+" f1="+f.format(evaluationResults.getF1()));
//		return evaluationResults;
//	}
	
	// Tagger evaluation for NER task
	public EvaluationResults evaluateTagger(String taggerName, TreeMap<String, InstanceList> taggerInstancesMap, TreeMap<String, InstanceList> estimatedGroundTruth) {
		//System.out.print("[MultipleTaggers] 	Evaluating NER tagger: " + taggerName);
		
		int numCorrectTokens, totalTokens;
		int[] numTrueSegments, numPredictedSegments, numCorrectSegments;
		int allIndex = CRFWithMultipleTaggers.START_TAGS.length;
		numTrueSegments = new int[allIndex + 1];
		numPredictedSegments = new int[allIndex + 1];
		numCorrectSegments = new int[allIndex + 1];

		totalTokens = numCorrectTokens = 0;
		for (int n = 0; n < numTrueSegments.length; n++)
			numTrueSegments[n] = numPredictedSegments[n] = numCorrectSegments[n] = 0;
		
		for (String filename : taggerInstancesMap.keySet()) {
			InstanceList taggerInstances = taggerInstancesMap.get(filename);
			InstanceList estimatedGroundTruthInstances = estimatedGroundTruth.get(filename);
			if (taggerInstances == null || estimatedGroundTruthInstances == null || taggerInstances.size() != estimatedGroundTruthInstances.size()) {
                System.out.println("[MultipleTaggers]- SKIPPING EVAL");
                continue;
			}
			
			int trueStart, predStart; // -1 for non-start, otherwise index into segmentStartTag
			for (int i=0; i<taggerInstances.size(); i++) {
				TermExtractorInstance predInstance = (TermExtractorInstance) taggerInstances.get(i);
				TermExtractorInstance gtInstance = (TermExtractorInstance) estimatedGroundTruthInstances.get(i);
				ArrayList<String> tokens = predInstance.getTokens();
				ArrayList<String> predLabels = predInstance.getLabels();
				ArrayList<String> gtLabels = gtInstance.getLabels();
				if (predLabels.size() != gtLabels.size()) {
					continue; // ok, ignore it... solutions??
				}
				
				for (int j=0; j<tokens.size(); j++) {
					totalTokens++;
					if (gtLabels.get(j).equals(predLabels.get(j)))
						numCorrectTokens++;
					trueStart = predStart = -1;
					// Count true segment starts
					for (int n = 0; n < CRFWithMultipleTaggers.START_TAGS.length; n++) {
						if (CRFWithMultipleTaggers.START_TAGS[n].equals(gtLabels.get(j))) {
							numTrueSegments[n]++;
							numTrueSegments[allIndex]++;
							trueStart = n;
							break;
						}
					}
					// Count predicted segment starts
					for (int n = 0; n < CRFWithMultipleTaggers.START_TAGS.length; n++) {
						if (CRFWithMultipleTaggers.START_TAGS[n].equals(predLabels.get(j))) {
							numPredictedSegments[n]++;
							numPredictedSegments[allIndex]++;
							predStart = n;
						}
					}
					if (trueStart != -1 && trueStart == predStart) {
						// Truth and Prediction both agree that the same segment
						// tag-type is starting now
						int m;
						boolean trueContinue = false;
						boolean predContinue = false;
						for (m = j + 1; m < gtLabels.size(); m++) {
							trueContinue = CRFWithMultipleTaggers.CONTINUE_TAGS[predStart]
									.equals(gtLabels.get(m));
							predContinue = CRFWithMultipleTaggers.CONTINUE_TAGS[predStart]
									.equals(predLabels.get(m));
							if (!trueContinue || !predContinue) {
								if (trueContinue == predContinue) {
									// They agree about a segment is ending somehow
									numCorrectSegments[predStart]++;
									numCorrectSegments[allIndex]++;
								}
								break;
							}
						}
						// for the case of the end of the sequence
						if (m == gtLabels.size()) {
							if (trueContinue == predContinue) {
								numCorrectSegments[predStart]++;
								numCorrectSegments[allIndex]++;
							}
						}
					}
				}
			}
		}
		
		// Compute precision, recall and f1
		double tokenAccuracy = ((double) numCorrectTokens) / ((double) totalTokens);
		DecimalFormat f = new DecimalFormat("0.####");
		double precision = numPredictedSegments[allIndex] == 0 ? 1
				: ((double) numCorrectSegments[allIndex]) / numPredictedSegments[allIndex];
		double recall = numTrueSegments[allIndex] == 0 ? 1
				: ((double) numCorrectSegments[allIndex]) / numTrueSegments[allIndex];
		double f1 = recall + precision == 0.0 ? 0.0
				: (2.0 * recall * precision) / (recall + precision);
		
		// Return evalutation results
		EvaluationResults evaluationResults = new EvaluationResults(precision, recall, f1);
		evaluationResults.setTokenAccuracy(tokenAccuracy);
		evaluationResults.setNumTokens(totalTokens);
		System.out.println(" -> tokenAcc="+f.format(tokenAccuracy)+" precision="+f.format(evaluationResults.getPrecision())+
				" recall="+f.format(evaluationResults.getRecall())+" f1="+f.format(evaluationResults.getF1()));
		return evaluationResults;
	}
}
