package pt.uc.dei.crf_ma.clustering;

public class DistanceMetrics {
	
	public static double euclideanDistance(double[] v1, double[] v2) { 
		assert(v1.length == v2.length);
		double distance = 0;
		for (int i = 0; i < v1.length; i++) {
			distance = distance + Math.pow((v1[i] - v2[i]), 2);
		}
		return Math.sqrt(distance);
	}
	
	public static double diceCoefficient(double[] v1, double[] v2) {
		assert(v1.length == v2.length);
		double intersection = 0;
		double cardinality1 = 0;
		double cardinality2 = 0;
		for (int i=0; i<v1.length; i++) {
			if (v1[i] == 1)
				cardinality1++;
			if (v2[i] == 1)
				cardinality2++;
			if (v1[i] == 1 && v2[i] == 1)
				intersection++;
		}
		return (2*intersection) / (cardinality1 + cardinality2);
	}
	
	public static double cosineDistance(double[] v1, double[] v2) {
		assert(v1.length == v2.length);
		double v1norm = 0;
		double v2norm = 0;
		double dot = 0;
		for (int i=0; i<v1.length; i++) {
			v1norm += v1[i]*v1[i];
			v2norm += v2[i]*v2[i];
			dot += v1[i]*v2[i];
		}
		v1norm = Math.sqrt(v1norm);
		v2norm = Math.sqrt(v2norm);
		return dot / (v1norm * v2norm);
	}
	
	public static double hammingDistance(double[] v1, double[] v2) {
		assert(v1.length == v2.length);
		double hamming = 0; 
		for (int i=0; i<v1.length; i++) {
			if (v1[i] != v2[i])
				hamming++;
		}
		return hamming;
	}
	
	public static double jaccardCoefficient(double[] v1, double[] v2) {
		assert(v1.length == v2.length);
		double intersection = 0.0;
		double union = v1.length;
		for (int i=0; i<v1.length; i++) {
			if (v1[i] == 1 && v2[i] == 1)
				intersection++;
		}
		return intersection / union;
	}
	
	/**
	 * Check if a distance matrix obeys the triangle inequity, i.e.
	 * dist(a,b) + dist(b,c) >= dist(a,c) for any a,b,c
	 * 
	 * @param matrix
	 * @return the number of violations of the inequity
	 */
	public static int checkTriangleInequity(double[][] matrix) {
		assert(matrix.length > 0 && matrix[0].length > 0); // make sure matrix is not empty
		assert(matrix.length == matrix[0].length); // make sure matrix is square
		int inequityViolations = 0;
		for (int a=0; a<matrix.length; a++) {
			for (int b=0; b<matrix.length; b++) {
				if (b != a) {
					for (int c=0; c<matrix.length; c++) {
						if (c != a && c != b) {
							if (matrix[a][b] + matrix[b][c] < matrix[a][c]) {
								inequityViolations++;
							}
						}
					}
				}
			}
		}
		return inequityViolations;
	}
}
