/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package pt.uc.dei.crf_ma.pipe;

import cc.mallet.pipe.Pipe;
import cc.mallet.types.Instance;
import cc.mallet.types.LabelAlphabet;
import cc.mallet.types.TokenSequence;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Arrays;
import java.util.regex.Pattern;

import opennlp.tools.postag.POSModel;
import opennlp.tools.postag.POSTaggerME;

/**
 * NOTE: POSTagger pipe most go before the Sentence2TokenSequence pipe in order 
 * to have access to the whole instead of only one token (the whole sentence is
 * needed for the part-of-speech tagging)
 *
 * @author fmpr
 */
public class POSTagger extends Pipe implements Serializable {


	private static final long serialVersionUID = 1L;
	private static final String DEFAULT_MODEL = "/models/opennlp/en-pos-maxent.bin";
	
	private POSTaggerME posTagger;
	private String prefix;
	private String modelpath;
	private boolean articlesAsInstances = false; // each instance is an article?

	public POSTagger(String prefix) {
		this(prefix, DEFAULT_MODEL);
	}
	
    public POSTagger(String prefix, String modelpath) {

        this.prefix = prefix;
        this.modelpath = modelpath;
        
        this.loadModel();
    }
    
    // Load OpenNLP POSTagger models
    private void loadModel() {
        InputStream modelIn = null;
        try {
            modelIn = getClass().getResourceAsStream(modelpath);
            POSModel model = new POSModel(modelIn);
            posTagger = new POSTaggerME(model);
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            if (modelIn != null) {
                try {
                    modelIn.close();
                } catch (IOException e) {
                	e.printStackTrace();
                }
            }
        }
    }

    @Override
    public Instance pipe(Instance carrier) {
        TokenSequence ts = (TokenSequence) carrier.getData();
        String[] tokens = new String[ts.size()];
        
        for (int i=0;i<ts.size();i++)
        	tokens[i] = ts.get(i).getText().toString();
        
        if(articlesAsInstances) {
	        int ind = 0;
	        for (int i=1;i<ts.size();i++) {
	        	if (ts.get(i).hasProperty(AnnotatedText2TokenSequence.START_SENTENCE_FEATURE)) {
	        		String [] tmptks = Arrays.copyOfRange(tokens, ind, i);
	        		
	        		String[] posTags = posTagger.tag(tmptks);
	        		for (int j=ind;j<=i;j++)
	        			ts.get(j).setFeatureValue(prefix+posTags[j], 1.0);
	        		
	        		ind = i+1;
	        	}
	        }
        }
        
        // Use OpenNLP to find the POS tags
        String[] posTags = posTagger.tag(tokens);
        for(int i = 0; i<ts.size(); i++)
        	ts.get(i).setFeatureValue(prefix+posTags[i], 1.0);
      
        return carrier;
    }

	// Serialization 
	
	private void writeObject (ObjectOutputStream out) throws IOException {
		out.writeLong(serialVersionUID);
		out.writeObject(prefix);
		out.writeObject(modelpath);
	}
	
	private void readObject (ObjectInputStream in) throws IOException, ClassNotFoundException {
		long version = in.readLong ();
		// we should check here if this is the correct version
		
		prefix = (String) in.readObject();
		modelpath = (String) in.readObject();
		loadModel();
	}
}
